optimize panning, and some cleanup
[lantea.git] / js / map.js
index ea3b62256c5d54ce23511f004445afc600ef0c73..a8105ae48f0bbb0845162a32ec85bc97e2b655dd 100644 (file)
--- a/js/map.js
+++ b/js/map.js
@@ -3,7 +3,7 @@
  * You can obtain one at http://mozilla.org/MPL/2.0/. */
 
 var gMapCanvas, gMapContext, gTrackCanvas, gTrackContext, gGeolocation;
-var gDebug = true;
+var gDebug = false;
 
 var gTileSize = 256;
 var gMaxZoom = 18; // The minimum is 0.
@@ -56,9 +56,6 @@ var gLastMouseX = 0;
 var gLastMouseY = 0;
 var gZoomFactor;
 
-// Used as an associative array.
-// The keys have to be strings, ours will be "xindex,yindex,zindex" e.g. "13,245,12".
-var gTiles = {};
 var gLoadingTile;
 
 var gMapPrefsLoaded = false;
@@ -122,7 +119,7 @@ function initMap() {
       gWaitCounter++;
       gTrackStore.getList(function(aTPoints) {
         if (gDebug)
-          document.getElementById("debug").textContent = aTPoints.length + " points loaded.";
+          console.log(aTPoints.length + " points loaded.");
         if (aTPoints.length) {
           gTrack = aTPoints;
         }
@@ -133,7 +130,7 @@ function initMap() {
       setTimeout(getPersistentPrefs, 100);
     loopCnt++;
     if (loopCnt > 50) {
-      document.getElementById("debug").textContent = "Loading prefs failed.";
+      console.log("Loading prefs failed.");
     }
   };
   getPersistentPrefs();
@@ -165,15 +162,19 @@ function initMap() {
 function resizeAndDraw() {
   var viewportWidth = Math.min(window.innerWidth, window.outerWidth);
   var viewportHeight = Math.min(window.innerHeight, window.outerHeight);
-
-  gMapCanvas.width = viewportWidth;
-  gMapCanvas.height = viewportHeight;
-  gTrackCanvas.width = viewportWidth;
-  gTrackCanvas.height = viewportHeight;
-  drawMap();
-  showUI();
+  if (gMapCanvas && gTrackCanvas) {
+    gMapCanvas.width = viewportWidth;
+    gMapCanvas.height = viewportHeight;
+    gTrackCanvas.width = viewportWidth;
+    gTrackCanvas.height = viewportHeight;
+    drawMap();
+    showUI();
+  }
 }
 
+// Using scale(x, y) together with drawing old data on scaled canvas would be an improvement for zooming.
+// See https://developer.mozilla.org/en-US/docs/Canvas_tutorial/Transformations#Scaling
+
 function zoomIn() {
   if (gPos.z < gMaxZoom) {
     gPos.z++;
@@ -210,7 +211,6 @@ function setMapStyle() {
   var mapSel = document.getElementById("mapSelector");
   if (mapSel.selectedIndex >= 0 && gActiveMap != mapSel.value) {
     gActiveMap = mapSel.value;
-    gTiles = {};
     document.getElementById("copyright").innerHTML =
         gMapStyles[gActiveMap].copyright;
     showUI();
@@ -259,12 +259,17 @@ function decodeIndex(encodedIdx) {
   return {x: ind[0], y: ind[1], z: ind[2]};
 }
 
-function drawMap() {
-  // Go through all the currently loaded tiles. If we don't want any of them remove them.
-  // for (t in gTiles) {
-  //   if (isOutsideWindow(t))
-  //     delete gTiles[t];
-  // }
+function drawMap(aPixels, aOverdraw) {
+  // aPixels is an object with left/right/top/bottom members telling how many
+  //   pixels on the borders should actually be drawn.
+  // aOverdraw is a bool that tells if we should draw placeholders or draw
+  //   straight over the existing content.
+  if (!aPixels)
+    aPixels = {left: gMapCanvas.width, right: gMapCanvas.width,
+               top: gMapCanvas.height, bottom: gMapCanvas.height};
+  if (!aOverdraw)
+    aOverdraw = false;
+
   document.getElementById("zoomLevel").textContent = gPos.z;
   gZoomFactor = Math.pow(2, gMaxZoom - gPos.z);
   var wid = gMapCanvas.width * gZoomFactor; // Width in level 18 pixels.
@@ -279,32 +284,48 @@ function drawMap() {
   if (gMapPrefsLoaded && mainDB)
     gPrefs.set("position", gPos);
 
-  // Go through all the tiles we want.
-  // If any of them aren't loaded or being loaded, do so.
+  var tiles = {left: Math.ceil((xMin + aPixels.left * gZoomFactor) / size) -
+                               (aPixels.left ? 0 : 1),
+               right: Math.floor((xMax - aPixels.right * gZoomFactor) / size) -
+                                 (aPixels.right ? 1 : 0),
+               top: Math.ceil((yMin + aPixels.top * gZoomFactor) / size) -
+                              (aPixels.top ? 0 : 1),
+               bottom: Math.floor((yMax - aPixels.bottom * gZoomFactor) / size) -
+                                  (aPixels.bottom ? 1 : 0)};
+
+  // Go through all the tiles in the map, find out if to draw them and do so.
   for (var x = Math.floor(xMin / size); x < Math.ceil(xMax / size); x++) {
     for (var y = Math.floor(yMin / size); y < Math.ceil(yMax / size); y++) { // slow script warnings on the tablet appear here!
-      // Round here is **CRUCIAL** otherwise the images are filtered
-      // and the performance sucks (more than expected).
-      var xoff = Math.round((x * size - xMin) / gZoomFactor);
-      var yoff = Math.round((y * size - yMin) / gZoomFactor);
-      // Draw placeholder, and then initiate loading/drawing of real one.
-      gMapContext.drawImage(gLoadingTile, xoff, yoff);
-
-      gTileService.get(gActiveMap, {x: x, y: y, z: gPos.z}, function(aImage, aStyle, aCoords) {
-        // Only draw if this applies for the current view.
-        if ((aStyle == gActiveMap) && (aCoords.z == gPos.z)) {
-          var ixMin = gPos.x - wid / 2;
-          var iyMin = gPos.y - ht / 2;
-          var ixoff = Math.round((aCoords.x * size - ixMin) / gZoomFactor);
-          var iyoff = Math.round((aCoords.y * size - iyMin) / gZoomFactor);
-          var URL = window.URL;
-          var imgURL = URL.createObjectURL(aImage);
-          var imgObj = new Image();
-          imgObj.src = imgURL;
-          gMapContext.drawImage(imgObj, ixoff, iyoff);
-          URL.revokeObjectURL(imgURL);
-        }
-      });
+      // Only go to the drawing step if we need to draw this tile.
+      if (x < tiles.left || x > tiles.right || y < tiles.top || y > tiles.bottom) {
+        // Round here is **CRUCIAL** otherwise the images are filtered
+        // and the performance sucks (more than expected).
+        var xoff = Math.round((x * size - xMin) / gZoomFactor);
+        var yoff = Math.round((y * size - yMin) / gZoomFactor);
+        // Draw placeholder tile unless we overdraw.
+        if (!aOverdraw)
+          gMapContext.drawImage(gLoadingTile, xoff, yoff);
+
+        // Initiate loading/drawing of the actual tile.
+        gTileService.get(gActiveMap, {x: x, y: y, z: gPos.z},
+                         function(aImage, aStyle, aCoords) {
+          // Only draw if this applies for the current view.
+          if ((aStyle == gActiveMap) && (aCoords.z == gPos.z)) {
+            var ixMin = gPos.x - wid / 2;
+            var iyMin = gPos.y - ht / 2;
+            var ixoff = Math.round((aCoords.x * size - ixMin) / gZoomFactor);
+            var iyoff = Math.round((aCoords.y * size - iyMin) / gZoomFactor);
+            var URL = window.URL;
+            var imgURL = URL.createObjectURL(aImage);
+            var imgObj = new Image();
+            imgObj.src = imgURL;
+            imgObj.onload = function() {
+              gMapContext.drawImage(imgObj, ixoff, iyoff);
+              URL.revokeObjectURL(imgURL);
+            }
+          }
+        });
+      }
     }
   }
   gLastDrawnPoint = null;
@@ -437,7 +458,13 @@ var mapEvHandler = {
           var dY = y - gLastMouseY;
           gPos.x -= dX * gZoomFactor;
           gPos.y -= dY * gZoomFactor;
-          drawMap();
+          var mapData = gMapContext.getImageData(0, 0, gMapCanvas.width, gMapCanvas.height);
+          gMapContext.clearRect(0, 0, gMapCanvas.width, gMapCanvas.height);
+          gMapContext.putImageData(mapData, dX, dY);
+          drawMap({left: (dX > 0) ? dX : 0,
+                   right: (dX < 0) ? -dX : 0,
+                   top: (dY > 0) ? dY : 0,
+                   bottom: (dY < 0) ? -dY : 0});
           showUI();
         }
         gLastMouseX = x;
@@ -471,10 +498,9 @@ var mapEvHandler = {
                        y: gPos.y + (x - gMapCanvas.height / 2) * gZoomFactor};
         var gpsCoord = xy2gps(ptCoord.x, ptCoord.y);
         var pt2Coord = gps2xy(gpsCoord.latitude, gpsCoord.longitude);
-        document.getElementById("debug").textContent =
-            ptCoord.x + "/" + ptCoord.y + " - " +
-            gpsCoord.latitude + "/" + gpsCoord.longitude + " - " +
-            pt2Coord.x + "/" + pt2Coord.y;
+        console.log(ptCoord.x + "/" + ptCoord.y + " - " +
+                    gpsCoord.latitude + "/" + gpsCoord.longitude + " - " +
+                    pt2Coord.x + "/" + pt2Coord.y);
         */
 
         var newZoomLevel = gPos.z + (delta > 0 ? 1 : -1);
@@ -587,7 +613,8 @@ function startTracking() {
       function(error) {
         // Ignore erros for the moment, but this is good for debugging.
         // See https://developer.mozilla.org/en/Using_geolocation#Handling_errors
-        document.getElementById("debug").textContent = error.message;
+        if (gDebug)
+          console.log(error.message);
       },
       {enableHighAccuracy: true}
     );